#include "_lib_sensors.h"
// START OF PRIVATE DEFINITIONS
//--------------------------------------------------------

// MUX select for IR sensor is as follow:
// PJ.1, PJ.0, P1.6
// 0 0 0 -> forward
// 0 0 1 -> reverse
// 0 1 0 -> right
// 0 1 1 -> left
// 1 0 0 -> down

#define MUX_IR_PIN_DIR_0 PJDIR
#define MUX_IR_PIN_DIR_1 PJDIR
#define MUX_IR_PIN_DIR_2 P1DIR

#define MUX_IR_PIN_OUT_0 PJOUT
#define MUX_IR_PIN_OUT_1 PJOUT
#define MUX_IR_PIN_OUT_2 P1OUT

#define MUX_IR_PIN_0_MASK 2
#define MUX_IR_PIN_1_MASK 1
#define MUX_IR_PIN_2_MASK 64


// IR photodiode result is transmitted to P6.0 (A0)
// [IR_RCV1_LO]
// Note: IR sensor result is NOT high-pass filtered
#define PIN_IR_RESULT_OUT P6OUT
#define PIN_IR_RESULT_DIR P6DIR
#define PIN_IR_RESULT_SEL P6SEL
#define PIN_IR_MASK 1

// This is the TX signal to drive the IR sensor
#define PIN_IR_TX_DIR P2DIR
#define PIN_IR_TX_OUT P2OUT
#define PIN_IR_TX_MASK 1



// Variables to hold ADC10 samplings
volatile uint16_t _num_adc10_samples = 0;
uint16_t ADC10_SAMPLE_TARGET = 20;
volatile uint32_t _accumulated_adc10 = 0;
volatile uint8_t _sampling_finished = 0;
volatile uint32_t sample;

#pragma vector = ADC10_VECTOR
__interrupt void ADC10_ISR(void) {
        sample = ADC10MEM0;
        _num_adc10_samples++;

        // We've collected all the samples
        if(_num_adc10_samples == ADC10_SAMPLE_TARGET){
                _num_adc10_samples = 0;
                // Disable ADC10
                ADC10IE &= ~ADC10IE0;
                // This is used by sensor_IR_poll() to block on
                // until the sampling is done
                _sampling_finished = 1;
                return;
        }

        // Alternately adding and subtracting
        _accumulated_adc10 += sample;

        // Enable conversion
        ADC10CTL0 |= ADC10SC;
}


//--------------------------------------------------------

// START OF PUBLIC DEFINITION
//--------------------------------------------------------

// Poll the specified IR sensor
// Direction:
// 0 -> forward
// 1 -> reverse
// 2 -> right
// 3 -> left
// 4 -> down
int sensor_IR_poll(uint8_t direction){
        // Return value
        uint32_t sensor_value;

        switch(direction){
                case 0:
                // Decode the MUX
                MUX_IR_PIN_OUT_0 &= ~MUX_IR_PIN_0_MASK;
                MUX_IR_PIN_OUT_1 &= ~MUX_IR_PIN_1_MASK;
                MUX_IR_PIN_OUT_2 &= ~MUX_IR_PIN_2_MASK;
                break;

                case 1:
                // Decode the MUX
                MUX_IR_PIN_OUT_0 &= ~MUX_IR_PIN_0_MASK;
                MUX_IR_PIN_OUT_1 &= ~MUX_IR_PIN_1_MASK;
                MUX_IR_PIN_OUT_2 |= MUX_IR_PIN_2_MASK;

                break;
                case 2:
                // Decode the MUX
                MUX_IR_PIN_OUT_0 &= ~MUX_IR_PIN_0_MASK;
                MUX_IR_PIN_OUT_1 |= MUX_IR_PIN_1_MASK;
                MUX_IR_PIN_OUT_2 &= ~MUX_IR_PIN_2_MASK;

                break;
                case 3:
                // Decode the MUX
                MUX_IR_PIN_OUT_0 &= ~MUX_IR_PIN_0_MASK;
                MUX_IR_PIN_OUT_1 |= MUX_IR_PIN_1_MASK;
                MUX_IR_PIN_OUT_2 |= MUX_IR_PIN_2_MASK;

                break;
                case 4:
                // Decode the MUX
                MUX_IR_PIN_OUT_0 |= MUX_IR_PIN_0_MASK;
                MUX_IR_PIN_OUT_1 &= ~MUX_IR_PIN_1_MASK;
                MUX_IR_PIN_OUT_2 &= ~MUX_IR_PIN_2_MASK;
                break;

                default:
                return -1;

        }
        // Enable the transmitter
        PIN_IR_TX_OUT |= PIN_IR_TX_MASK;

        __delay_cycles(7000);

        // Enable ADC10
        ADC10CTL0 |= ADC10SC;

        // Reset sensor value and state
        _sampling_finished = 0;
        _accumulated_adc10 = 0;

        // Enable ADC10
        ADC10IE |= ADC10IE0;

        // Wait for samples to be collected
        while(!_sampling_finished){
        }

        sensor_value = _accumulated_adc10;

        // Disable IR emitter to save power
        PIN_IR_TX_OUT &= ~PIN_IR_TX_MASK;

        return sensor_value;
}

void sensor_Initialize(){
        // Set the MUX select PINS for IR as output PINS
        MUX_IR_PIN_DIR_0 |=  MUX_IR_PIN_0_MASK;
        MUX_IR_PIN_DIR_1 |=  MUX_IR_PIN_1_MASK;
        MUX_IR_PIN_DIR_2 |=  MUX_IR_PIN_2_MASK;

        // Enable bit of the IR emitter
        PIN_IR_TX_DIR |= PIN_IR_TX_MASK;

        // Set P6.0 to be A0 channel for ADC10
        PIN_IR_RESULT_SEL |= PIN_IR_MASK;

        // ADC10
        ADC10CTL0 &= ~ADC10ENC;                   // Disable conv to allow mod of ADC10CTL0 0xFFFx
        ADC10CTL0 = ADC10SHT_4 + ADC10ON;           // turn on ADC10, S&H 32 CLKs
        ADC10CTL1 = ADC10SHP/* + ADC10SSEL_3/* + ADC10CONSEQ_2*/;                       // use sampling timer
        ADC10IE |= ADC10IE0;                      // Enable Interrupt
        ADC10CTL0 |= ADC10ENC;          // Enable conversions

        __bis_SR_register(GIE); // Global interrupt enabled
}


//--------------------------------------------------------
