// System-level library, such as GPIOs initialization or watchdog controll
// 
// Author: Antonius Denny Harijanto
// 

#include "_lib_system.h"
#include <stdint.h>

// ----------------------------------------------------------------------
// Lower level motor control, not needed to be known by the Blimp program
// ----------------------------------------------------------------------
#define VMotor_H 1 // P 1.0
#define VMotor_L 2 // P 1.1
#define LMotor_H 4 // P 1.2
#define LMotor_L 8 // P 1.3
#define RMotor_H 16 // P 1.4
#define RMotor_L 32 // P 1.5

// Belows are the macro that calls sys_Motor()
// Literally runs the motor with full speed
//
// Speed control is implemented using PWM and is exposed in
// "Higher level motor control" section above.
//
// Values of HI and LO pins of the motor correspond to the following:
//
// Standby : 0 0
// Reverse : 1 0 
// Forward : 0 1
// Brake : 0 0
#define __sys_VMotor_standby sys_Motor(VMotor_H, VMotor_L, 1, 1)
#define __sys_RMotor_standby sys_Motor(RMotor_H, RMotor_L, 1, 1)
#define __sys_LMotor_standby sys_Motor(LMotor_H, LMotor_L, 1, 1)

#define __sys_VMotor_reverse sys_Motor(VMotor_H, VMotor_L, 1, 0)
#define __sys_RMotor_reverse sys_Motor(RMotor_H, RMotor_L, 1, 0)
#define __sys_LMotor_reverse sys_Motor(LMotor_H, LMotor_L, 1, 0)

#define __sys_VMotor_forward sys_Motor(VMotor_H, VMotor_L, 0, 1)
#define __sys_RMotor_forward sys_Motor(RMotor_H, RMotor_L, 0, 1)
#define __sys_LMotor_forward sys_Motor(LMotor_H, LMotor_L, 0, 1)

#define __sys_VMotor_brake sys_Motor(VMotor_H, VMotor_L, 0, 0)
#define __sys_RMotor_brake sys_Motor(RMotor_H, RMotor_L, 0, 0)
#define __sys_LMotor_brake sys_Motor(LMotor_H, LMotor_L, 0, 0)

#define __LMotor_CCR_Forward TA0CCR1
#define __LMotor_CCR_Reverse TA0CCR2
#define __RMotor_CCR_Forward TA0CCR3
#define __RMotor_CCR_Reverse TA0CCR4

// Direction of the vertical motor
// Direction = 1 -> Forward
// Direction = 0 -> Reverse
static uint8_t VMotor_direction;

// Control the motor using this convention:
//
// Pin1 and Pin2 are the HI and LO pins of the motor,
// in1 and in2 are the values for the pins.
static void sys_Motor(uint8_t pin1, uint8_t pin2, uint8_t in1, uint8_t in2){
	if (in1){
		P1OUT |= pin1;
	}
	else {
		P1OUT &= ~pin1;
	}

	if (in2) {
		P1OUT |= pin2;
	}
	else {
		P1OUT &= ~pin2;
	}
}

// Helper function to be called by sys_Initialize
static void sys_Initialize_Motor(){
	// Set the pins as output
	P1DIR |= VMotor_H;
	P1DIR |= VMotor_L;
	P1DIR |= RMotor_H;
	P1DIR |= RMotor_L;
	P1DIR |= LMotor_H;
	P1DIR |= LMotor_L;	

	// Set the option select
	// So that the pins could be used for PWM output
	//P1SEL |= VMotor_H;
	//P1SEL |= VMotor_L;
	P1SEL |= RMotor_H;
	P1SEL |= RMotor_L;
	P1SEL |= LMotor_H;
	P1SEL |= LMotor_L;

	// Put all motors into standby
	sys_Motor(VMotor_H, VMotor_L, 1, 1);
	sys_Motor(RMotor_H, RMotor_L, 1, 1);
	sys_Motor(LMotor_H, LMotor_L, 1, 1);

	// Set up TIMERA for motor control PWM
	TA0CCR0 = MOTOR_PWM_PERIOD;
	// Disable any motor by default
	TA0CCR1 = 0;
	TA0CCR2 = 0;
	TA0CCR3 = 0;
	TA0CCR4 = 0;

	TA0CCTL1 = OUTMOD_7;
	TA0CCTL2 = OUTMOD_7;

	TA0CCTL3 = OUTMOD_7;
	TA0CCTL4 = OUTMOD_7;

	// Use SMCLK, Up to CCR0, clear the timer
	TA0CTL = TASSEL_2 + MC_1 + TACLR;

	__bis_SR_register(GIE); // Global interrupt enabled
}


static void sys_VMotor_run(uint32_t freq, uint8_t direction){
	// Disable the timer
	TB0CCTL1 &= ~CCIE;
	TB0CCTL0 &= ~CCIE;
	// try to stop the vertical fan
	if (freq == 0){
		TB0CCR0 = 0;
		TB0CCR1 = 0;
		sys_Motor(VMotor_L,VMotor_H, 1,1);
		return;
	}


	// Stop the motor	
	__sys_VMotor_standby;

	// Set the direction that is being used by the timer
	VMotor_direction = direction;

	// 10 is the constant known to work
	TB0CCR0 = MOTOR_PWM_PERIOD;
	TB0CCR1 = freq << (MOTOR_PWM_PERIOD_SHIFT - 8);

	// Enable timer interrupt
	TB0CCTL1 |= CCIE;
	TB0CCTL0 |= CCIE;

	// Enable timer B as a whole
	TB0CTL = TBSSEL_2 + MC_1 + TBCLR + TBIE;

}

// TODO: These two need to be implemented using software PWM
void sys_VMotor_forward(uint32_t freq){
	sys_VMotor_run(freq, 1);
}

void sys_VMotor_reverse(uint32_t freq){
	sys_VMotor_run(freq, 0);
}

void sys_LMotor_forward(uint32_t freq) {
	__sys_LMotor_standby;
	__LMotor_CCR_Reverse = 0;
	__LMotor_CCR_Forward = freq << (MOTOR_PWM_PERIOD_SHIFT - 8);
}

void sys_LMotor_reverse(uint32_t freq) {
	__sys_LMotor_standby;
	__LMotor_CCR_Forward = 0;
	__LMotor_CCR_Reverse = freq << (MOTOR_PWM_PERIOD_SHIFT - 8);
}

void sys_RMotor_forward(uint32_t freq) {
	__sys_RMotor_standby;
	__RMotor_CCR_Reverse = 0;
	__RMotor_CCR_Forward = freq << (MOTOR_PWM_PERIOD_SHIFT - 8);
}

void sys_RMotor_reverse(uint32_t freq) {
	__sys_RMotor_standby;
	__RMotor_CCR_Forward = 0;
	__RMotor_CCR_Reverse = freq << (MOTOR_PWM_PERIOD_SHIFT - 8);
}

void sys_Initialize() {
	// Disable the Watchdog Timer
	WDTCTL = WDTPW + WDTHOLD;
	// Setup CPU Clock to be 12MHz
	// Setup UCS: ACLK = REFO; MCLK = SMCLK = 12MHz                                               
	UCSCTL3 |= SELREF_2;                      // Set DCO FLL reference = REFO
	UCSCTL4 |= SELA_2;                        // Set MCLK = SMCLK = DCOCLK
	__bis_SR_register(SCG0);                  // Disable the FLL control loop
	UCSCTL0 = 0x0000;                         // Set lowest possible DCOx, MODx
	UCSCTL1 = DCORSEL_5;                      // Select DCO range 24MHz operation
	UCSCTL2 = FLLD_1 + 374;                   // Set DCO Multiplier for 12MHz
	// (N + 1) * FLLRef = Fdco
	// (374 + 1) * 32768 = 12MHz
	// Set FLL Div = fDCOCLK/2
	__bic_SR_register(SCG0);                  // Enable the FLL control loop


	// Initializes LED and turn off all of them by default
	RED_LED_SETUP;
	GREEN_LED_SETUP;

	// Indicate that the system is ready
	GREEN_LED_ON;

	// Initialize motors
	// --------------------------------------------------------------
	sys_Initialize_Motor();
	// --------------------------------------------------------------

	__bis_SR_register(GIE);                   // Global interrupts enabled
}

#pragma vector=TIMERB0_VECTOR
__interrupt void TIMERB0_ISR(void){

	if(VMotor_direction == 1){
		__sys_VMotor_forward;
	}
	else if(VMotor_direction == 0){
		__sys_VMotor_reverse;
	}
}

// Timer_B7 Interrupt Vector (TBIV) handler                                                                                                                   
#pragma vector=TIMERB1_VECTOR
__interrupt void TIMERB1_ISR(void)
{
	/* Any access, read or write, of the TBIV register automatically resets the
  highest "pending" interrupt flag. */
	switch(TBIV)
	{
	case  0:                           // No Interrupt
		break;
	case  2:                           // CCR1 used
		__sys_VMotor_standby;
		break;
	default: break;
	}
}
