// Implements the UART connectivity between the Blimp and the wifi board
//
// Author: Antonius Denny Harijanto
//

// Uncomment below to enable UART debugging
// #define UART_DEBUG
#include "_msp_uart_controller.h"
#include "../utils/queue.h"
#include "../wci/uart_to_wci.h"


// UART (ez430RF2500 <-> BlimpBot) interrupt routine
// This function handles the bit-level receiving and sending of data.
// This function calls the callback when a block of data is received from ez430
#ifndef UART_DEBUG

//-----------------------------------------------------------------------------------------
// Definitions
//-----------------------------------------------------------------------------------------
// The maximum data on the queue
#define TX_BUFFER_SIZE 20

// The value of UCA0IV indicates which UART interrupt 
// is being serviced
#define TX_ISR 4
#define RX_ISR 2


// A queue that holds the characters to be transmitted
QUEUE_DATA __transmit_queue;

// Circular buffer used by the queue
QUEUE_DATATYPE __transmit_buffer[TX_BUFFER_SIZE];
volatile char s [5] = {'\0'};
volatile unsigned int count = 0;

#pragma vector=USCI_A1_VECTOR
__interrupt void USCI_A1_ISR(void)
{
	// Temporary variable to hold data
	static char data;
	
	switch(UCA1IV)
	{
	case 0:
		break;

	case RX_ISR:
		data = UCA1RXBUF;
		// testing purpose
		s[count] = data;
		count++;
		if (count == 5){
			count = 0;
		}
		// Transfer the data to the WCI module
		uart_to_wci_copy(data);
		break;

	case TX_ISR:
		// Transmit the data if the data queue isn't empty
		if(queue_pop(&__transmit_queue, &data)){
			UCA1TXBUF = data;
		}
		// Case that should never be reached
		else{
			// Send some random data just so TXBUF won't block
			UCA1TXBUF = ';';
		}
		
		// If the data queue is empty, disable the TX interrupt
		if(queue_isEmpty(&__transmit_queue)){
			uart_Disable_TX_interrupt();
		}

		break;

	default:
		break;
	}
}
#endif


void uart_Initialize(void)
{       
	// Initializes the transmit buffer
	queue_create(&__transmit_queue, __transmit_buffer, TX_BUFFER_SIZE);

	// UCA1 is the one wired to P4.4 and 4.5
	P4SEL = BIT5+BIT4;                        // P4.5,4 = USCI_A0 TXD/RXD

	UCA1CTL1 |= UCSWRST;                      // **Put state machine in reset**
	UCA1CTL1 |= UCSSEL_2;                     // SMCLK
	UCA1CTL0 |= UCMSB;                        // Send data in a MSB-manner

	// N = (CLOCK_SPEED / BAUT_RATE) (See user's guide for more info)
	UCA1BR0 = 78;                             // This is oversampling rate, which is N / 16
	// UCBRF0 = 2 is the precision if the above has fraction
	// round([(N/16) - INT(N/16)] * 16)
	UCA1BR1 = 0;                              //
	UCA1MCTL = UCOS16 + UCBRF_2;              // UCOS16 is to indicate oversampling,  UCBRF_2 is explained above
	UCA1CTL1 &= ~UCSWRST;                     // **Initialize USCI state machine**

	UCA1IE |= UCRXIE;                         // Enable USCI_A0 interrupts

	__bis_SR_register(GIE);                   // Global interrupts enabled
}


void uart_Disable_RX_interrupt(void)
{
	UCA1IE &= ~UCRXIE;
}

void uart_Enable_RX_interrupt(void)
{
	UCA1IE |= UCRXIE;
}

void uart_Disable_TX_interrupt(void)
{
	UCA1IE &= ~UCTXIE;
}

void uart_Enable_TX_interrupt(void)
{
	UCA1IE |= UCTXIE;
}

uint8_t uart_Transmit_data(char *str, uint8_t length){
	while(length != 0){
		length--;
		while(!(UCA1IFG & UCTXIFG)){
		}
		UCA1TXBUF = *str++;
	}
	
//
//	// Return 0 if the queue isn't big enough to contain the entire data
//	if( (__transmit_queue.size - __transmit_queue.numElementsInQueue) < length){
//		uart_Enable_TX_interrupt();
//		return 0;
//	}
//
//	uint8_t i;
//	for (i = 0 ; i < length ; i++) {
//		// If the queue is full returns 0
//		queue_push(&__transmit_queue, *str++);
//	}
//
//	uart_Enable_TX_interrupt();
//
//	// Return success
//	return 1;
}


/*
---- DEBUGGING PART OF THE CODE ----
 */

// Debug UART as follow: Program the ez430 to keep sending 
// 'a' with baut rate 9600, if the RED led blinks that means
// the data is received. Also program the ez430 to blink
// when 'a' is received, if it blinks that means TX is working!
#ifdef UART_DEBUG
volatile unsigned int counter = 0;
volatile unsigned int counter2 = 0;
volatile char received;
// Echo back RXed character, confirm TX buffer is ready first
#pragma vector=USCI_A1_VECTOR
__interrupt void USCI_A1_ISR(void)
{
	switch(UCA1IV)
	{
	case 0:break;                             // Vector 0 - no interrupt
	case RX_ISR:                              // Vector 2 - RXIFG

		received = UCA1RXBUF;

		if(counter++ == 9600/10){              // Simple check to see if the timing is correct
			counter = 0;                        // If we got the baut rate correctly, LED should blink every 1 second
			RED_LED_TOGGLE;
		}

		break;
	case TX_ISR:
		if(counter2++ == 9600/10){              // Simple check to see if the timing is correct
			counter2 = 0;                        // If we got the baut rate correctly, LED should blink every 1 second
			GREEN_LED_TOGGLE;
		}
		UCA1TXBUF = 'a';

		break;                             // Vector 4 - TXIFG
	default: break;
	}
}
#endif
