// Matthew Kraus
// blimp_commands.c -- general structure of blimp commands
// module prefix: cmd_
// version 1.2
//
// Description
// When sending commands over the network, everything is in text format.
// For example, a string like this might come in over the network: $MOT1>128
// This string is "tokenized" by the Wireless Command Interface (WCI) module
// into the opcodes in this file. Additionally, each of the arguments to this
// command are tokenized into one-byte integers. (Some arguments are integer
// literals, like motor speed, that are not tokenized.) This tokenizing process
// can be thought of as converting the text command into a binary format
// that is programmer friendly (think: just like compilers do).
//
// These commands are documented in "Blimp Commands.pdf".

#pragma once

#ifndef INCLUDED_BLIMP_COMMANDS
#define INCLUDED_BLIMP_COMMANDS

#ifndef INCLUDED_MSP430_DATATYPES
	#include "msp430_datatypes.h"
#endif

//
// COMMAND FRAMES /////////////////////////////////////////
//
// Think of a "command frame" as a binary representation of
// text commands like $MOT1>128. Example:
// 
// CMD_COMMAND_FRAME cf;
// cf.OPcode = CMD_OPCODE_MOTOR;
// cf.arg[0] = CMD_MOT_ID1;
// cf.arg[1] = CMD_MOT_DIR_FWD;
// cf.arg[2] = 128;
//
// Notice that some arguments have tokenized representations
// that are #defined, and other arguments (like speed) are
// literal integers. See the arguments section for each command
// below to find out which arguments are tokens and which arguments
// are integer literals.
//
// A command frame has an opcode and a list of arguments.
// Some opcodes don't use arguments, others only use one byte.
// See the Blimp Instruction Set Architecture doc for more info.
//
#define CMD_MAX_NUM_ARGS 10
typedef struct {
	uint8_t OPcode;                 // an opcode is a blimp command (see opcode defines below)
	uint8_t arg[CMD_MAX_NUM_ARGS];  // arguments to the command
} CMD_COMMAND_FRAME, *CMD_pCOMMAND_FRAME;


//
// OPCODES ////////////////////////////////////////////////
//
// must be unique and <= 255; 
// some opcodes have multiple sub-functions, i.e. "STAT" can query system status or temperature
// depending on what arguments are passed. See the Blimp Instruction Set Architecture for more info.
#define CMD_OPCODE_NOP            0     // no operation, do nothing
#define CMD_OPCODE_RUN            1     // enable sensors and motors
#define CMD_OPCODE_STOP           2     // disable sensors and motors
#define CMD_OPCODE_RESET          3     // force reset
#define CMD_OPCODE_PING           4     // keepalive, ping
#define CMD_OPCODE_STAT           5     // return status from a query command
#define CMD_OPCODE_MOTOR          6     // Set{"Motor ID[0,1,2]", "Direction [Fwd,Rev]", "Duty Cycle[0-255]"}
#define CMD_OPCODE_SET            7     // TODO
#define CMD_OPCODE_TUNE           8     // TODO
#define CMD_OPCODE_POLL           9     // TODO
#define CMD_OPCODE_QUERY         10     // query motion controller status, system status, temperature, etc.2
#define CMD_OPCODE_PID           11     // turn PID on/off
#define CMD_OPCODE_CA            12     // turn collision avoidance on/off
#define CMD_OPCODE_TAR           13     // TARget sensor value for PID

//
// MOTOR ARGUMENTS //////////////////////////////////////////
//
// MOT arguments have three arguments, in this order:
// 1. ID (#define token)
// 2. DIR (#define token)
// 3. SPEED (integer literal)

#define CMD_MOT_ID0               '0'
#define CMD_MOT_ID1               '1'
#define CMD_MOT_ID2               '2'
							      
#define CMD_MOT_DIR_FWD           '>'
#define CMD_MOT_DIR_REV           '<'
#define CMD_MOT_DIR_UP            '^'
#define CMD_MOT_DIR_DOWN          'v'

//
// SYSTEM IDs (for QRY and STAT commands) /////////////////
//
// Both the query (QRY) and status (STAT) commands share
// the same system IDs (SYSID). Each system (or sub-system)
// is tokenized as a single byte character that can be
// evaluated as an 8-bit integer.

#define CMD_SYSID_SYSTEM          'S'
#define CMD_SYSID_MOTION          'M'
#define CMD_SYSID_TEMP            'T'

//
// STATUS ARGUMENTS ////////////////////////////////////////
//
// A STAT command has two arguments in this order:
// 1. SYSID (see above for system ID's)
// 2. STATE
// 
#define CMD_STAT_STATE_UP         0
#define CMD_STAT_STATE_DOWN       1
#define CMD_STAT_STATE_ERROR      2


// PID ARGUMENTS ////////////////////////////////////////////
#define CMD_PID_OFF               0
#define CMD_PID_ON                1

// CA (collision avoidance) ARGUMENTS ///////////////////////
#define CMD_CA_OFF                0
#define CMD_CA_ON                 1

// TAR CMD_OPCODE_TAR ///////////////////////////////////////
// argument: 2 character integer literal

#endif // INCLUDED_BLIMP_COMMANDS
