// Matthew Kraus
// fsm.c -- finite state machine implementation, optimized for embedded systems
// module prefix: fsm_
//

#include "fsm.h"

// assigns the starting state.
// No transition functions are executed, and neither is onEnter(). Only
// the state is assigned. This is not a normal state transition and should
// be used when starting the state machine or for reset scenarios
void fsm_init(FSM_STATE* currentState, FSM_STATE* startingState)
{
	currentState = startingState;
}

// internal use.
static void fsm_changeState(FSM_STATE* currentState, FSM_STATE* nextState, void* rawInput)
{
	if (nextState == NULL) return; // fail silently

	// found accepting state; do things needed to leave the state
	if (currentState->onExit != NULL) currentState->onExit(rawInput);

	// change state
	currentState = nextState;
				
	// here is where the state processes the input
	if (currentState->onEnter != NULL) {
		currentState->onEnter(rawInput);				
	}
}

// 'rawInput' might trigger a state change. find out if any of the
// outgoing states will accept this input, then call the following
// functions:
// 1. onExit (if defined)
// 2. onEnter (if defined)
//
// If no outgoing states are defined, then the state is switched to the
// defaultState.
void fsm_onInput(FSM_STATE* currentState, void* rawInput)
{
	uint8_t i;

	if (currentState == NULL) return; // invalid start state

	if (currentState->nextState != NULL) {
		// do any of the outgoing connected states accept this input?
		for (i = 0; i < currentState->numNextStates; i++) {
			if (currentState->nextState[i]->acceptsInput(rawInput) == TRUE) {
				// found accepting state, do transition
				fsm_changeState(currentState, currentState->nextState[i], rawInput);
				
				return;  // state change handled
			}
		} // end for
	}

	// transition to default state
	fsm_changeState(currentState, currentState->defaultState, rawInput);
}
