//
// queue.c
//

#include "queue.h"


//
// F U N C T I O N S //////////////////////////////////////////////////////////////////////////////
//


///////////////////////////////////////////////////////////////////////////////////////////////////
// Attach 'buffer' to this queue. The queue now manages access to 'buffer'.
// Note: for embedded systems, you might want to statically assign the buffer, as in this example:
//
// #define WCI_BUFFER_SIZE 16
// unsigned char WCI_BUFFER[WCI_BUFFER_SIZE];
//
// QUEUE_DATA wci_RXBUFFER = {
//    WCI_BUFFER,
//    0,
//    0,
//    0,
//    WCI_BUFFER_SIZE
//}; 
//
// If you statically assign the buffer as in the example above, there is no need to call
// this function to initialize the queue.
void queue_create(PQUEUE_DATA queueData, QUEUE_DATATYPE* buffer, uint8_t sizeOfBuffer)
{
	queueData->circularBuffer = buffer;
	queueData->head = 0;
	queueData->tail = 0;
	queueData->size = sizeOfBuffer;
	queueData->numElementsInQueue = 0;
}


///////////////////////////////////////////////////////////////////////////////////////////////////
//
BOOL queue_isEmpty(PQUEUE_DATA queueData)
{
	return queueData->numElementsInQueue == 0;
}


///////////////////////////////////////////////////////////////////////////////////////////////////
//
BOOL queue_isFull(PQUEUE_DATA queueData)
{
	return queueData->numElementsInQueue == queueData->size;
}


///////////////////////////////////////////////////////////////////////////////////////////////////
//
BOOL queue_push(PQUEUE_DATA queueData, QUEUE_DATATYPE val)
{
	if (queue_isFull(queueData) ) return FALSE;

	queueData->circularBuffer[queueData->tail++] = val;
	++queueData->numElementsInQueue;

	if (queueData->tail > queueData->size - 1) queueData->tail = 0;

	return TRUE;
}


///////////////////////////////////////////////////////////////////////////////////////////////////
//
BOOL queue_pop(PQUEUE_DATA queueData, QUEUE_DATATYPE* retVal)
{
	if (queue_isEmpty(queueData) ) return FALSE;

	*retVal = queueData->circularBuffer[queueData->head++];
	--queueData->numElementsInQueue;

	if (queueData->head > queueData->size - 1) queueData->head = 0;

	return TRUE;
}


///////////////////////////////////////////////////////////////////////////////////////////////////
// Reads the value at a given offset in the queue, but does not change the queue.
// Returns true if the read succeeded, false if the index was out of range
BOOL queue_peekAt(PQUEUE_DATA queueData, uint8_t idx, QUEUE_DATATYPE* retVal)
{	
	uint8_t offset;

	if (queue_isEmpty(queueData) ) return FALSE;
	if (idx >= queueData->numElementsInQueue || idx < 0) return FALSE;

	offset = queueData->head + idx;
	if (offset >= queueData->size) offset -= queueData->size;

	*retVal = queueData->circularBuffer[offset];

	return TRUE;
}


///////////////////////////////////////////////////////////////////////////////////////////////////
// Looks (peeks) ahead at the head of the queue to see if the values[] are in the queue.
// If all of the values in the null terminated array 'values' are found in the queue, true is returned.
// If the queue is empty, false is returned even if values is empty.
// 'values' must be null-terminated.
// similar to strcmp(), but returns a strictly true/false answer instead of a less than / greater than answer.
BOOL queue_startsWith(PQUEUE_DATA queueData, QUEUE_DATATYPE* values)
{
	uint8_t idx;
	uint8_t numLeft;
	QUEUE_DATATYPE* pValues;

	if (queue_isEmpty(queueData) || values == 0) return FALSE;

	idx     = queueData->head;
	numLeft = queueData->numElementsInQueue;
	pValues = values;

	while(*pValues != 0 && numLeft != 0) {
		if (queueData->circularBuffer[idx] != *pValues) return FALSE;
		
		--numLeft;
		++idx;
		if (idx > queueData->size - 1) idx = 0;
		
		++pValues;
	}

	return TRUE;
}

///////////////////////////////////////////////////////////////////////////////////////////////////
// returns true if 'val' is contained *anywhere* in the given queue
BOOL queue_contains(PQUEUE_DATA queueData, QUEUE_DATATYPE val)
{
	uint8_t idx;
	uint8_t numLeft;
	
	if (queue_isEmpty(queueData)) return FALSE;

	idx     = queueData->head;
	numLeft = queueData->numElementsInQueue;
	
	while (numLeft > 0) {
		if (queueData->circularBuffer[idx] == val) return TRUE;

		--numLeft;
		++idx;
		if (idx > queueData->size - 1) idx = 0;
	}

	return FALSE;
}
