// Matthew Kraus
// test_wci.c -- test wireless control interface
//

#include "uart_to_wci.h"
#include "../utils/queue.h"
#include "../utils/blimp_commands.h"
#include "wifi_ctrl_iface.h"

#include <stdio.h> // only for testing; not required on MSP430

CMD_COMMAND_FRAME cf;
int NUM_TESTS_FAILED = 0;

extern QUEUE_DATA wci_RXBUFFER; // for testing


///////////////////////////////////////////////////////////////////////////////
// Simulates sending data from the UART to the WCI module.
// Although this sends an entire string at once, the 
// uart_to_wci module sends only one char at a time. This
// send() is here for testing convenience.
//
// @arg fromUart must be null terminated
static void send(char* fromUart)
{
	char* pstr = fromUart;

	while(*pstr) {
		uart_to_wci_copy(*pstr);
		pstr++;
	}
} // end send()


///////////////////////////////////////////////////////////////////////////////
static void printCommand()
{
	// print opcode
	switch(cf.OPcode) {
		case CMD_OPCODE_NOP  : printf("CMD_OPCODE_NOP\n");   break;
		case CMD_OPCODE_RUN  : printf("CMD_OPCODE_RUN\n");   break;   
		case CMD_OPCODE_STOP : printf("CMD_OPCODE_STOP\n");  break;   
		case CMD_OPCODE_RESET: printf("CMD_OPCODE_RESET\n"); break;   
		case CMD_OPCODE_PING : printf("CMD_OPCODE_PING\n");  break;   
		case CMD_OPCODE_STAT | CMD_OPCODE_QUERY : 
			if (cf.OPcode == CMD_OPCODE_STAT)  printf("CMD_OPCODE_STAT\n");
			if (cf.OPcode == CMD_OPCODE_QUERY) printf("CMD_OPCODE_QUERY\n");
			
			// STAT and QRY share same system IDs (SYSID) for subsystems
			if (cf.arg[0] == CMD_SYSID_SYSTEM) printf("\tSYSID = CMD_SYSID_SYSTEM\n");
			if (cf.arg[0] == CMD_SYSID_MOTION) printf("\tSYSID = CMD_SYSID_MOTION\n");
			if (cf.arg[0] == CMD_SYSID_TEMP)   printf("\tSYSID = CMD_SYSID_TEMP\n");

			if (cf.OPcode == CMD_OPCODE_STAT) {
				// The STAT command has these STATE arguments (QRY command does NOT have this argument)
				if (cf.arg[1] == CMD_STAT_STATE_UP)    printf("\tSTATE = CMD_STAT_STATE_UP\n");
				if (cf.arg[1] == CMD_STAT_STATE_DOWN)  printf("\tSTATE = CMD_STAT_STATE_DOWN\n");
				if (cf.arg[1] == CMD_STAT_STATE_ERROR) printf("\tSTATE = CMD_STAT_STATE_ERROR\n");
			}
			break;   
		case CMD_OPCODE_MOTOR: 
			printf("CMD_OPCODE_MOTOR\n"); 
			if      (cf.arg[0] == CMD_MOT_ID0) printf("\tID = CMD_MOT_ID0\n");
			else if (cf.arg[0] == CMD_MOT_ID1) printf("\tID = CMD_MOT_ID1\n");
			else if (cf.arg[0] == CMD_MOT_ID2) printf("\tID = CMD_MOT_ID2\n");

			if      (cf.arg[1] == CMD_MOT_DIR_FWD)  printf("\tDIR = CMD_MOT_DIR_FWD\n");
			else if (cf.arg[1] == CMD_MOT_DIR_REV)  printf("\tDIR = CMD_MOT_DIR_REV\n");
			else if (cf.arg[1] == CMD_MOT_DIR_UP)   printf("\tDIR = CMD_MOT_DIR_UP\n");
			else if (cf.arg[1] == CMD_MOT_DIR_DOWN) printf("\tDIR = CMD_MOT_DIR_DOWN\n");

			printf("\tSPEED = %i\n", cf.arg[2]);
			break;   
		case CMD_OPCODE_SET  : printf("CMD_OPCODE_SET\n"); break;   
		case CMD_OPCODE_TUNE : printf("CMD_OPCODE_TUNE\n"); break;   
		case CMD_OPCODE_POLL : printf("CMD_OPCODE_POLL\n"); break;   
	}; // end switch

} // end printCommand()


///////////////////////////////////////////////////////////////////////////////
//
void printQueue()
{	
	char val;
	int i;
	printf("queue contains: ");

	for (i = 0; i < wci_RXBUFFER.numElementsInQueue; i++) {
		queue_peekAt(&wci_RXBUFFER, i, &val);
		printf("%c", val);
	}
	printf("\n");
}


///////////////////////////////////////////////////////////////////////////////
// Tests a given command string to see if the parser recognizes it.
// if 'expectSuccess' is true, then the commandString should be valid
// if 'expectSuccess' is false, then the commandString should be recognized as bad
//
void runTest(char* commandString, unsigned char expectSuccess, int expectedOPCODEor0, int line)
{
	// 1. print command in question
	printQueue();
	if (*commandString != 0) printf("Sending '%s'\n", commandString);
	else printf("(existing buffer data parsed)\n");
	
	// 2. simulate sending data from UART to blimp WCI module
	send(commandString);
	printQueue();

	// 3. analyze command
	if (wci_hasNext()) {		
		wci_nextCommand(&cf);	
		if (expectSuccess && cf.OPcode == expectedOPCODEor0) {
			printCommand();		
			printf("[PASS]\n");
		} else {
			printf("--->wrong command detected on [%s]:[%i]\n", __FILE__, line);
			printCommand();		
			printf("[FAIL]\n");
			++NUM_TESTS_FAILED;
		}
	} else if (expectSuccess) {
		printf("--->expected a command to be detected but didn't get one [%s]:[%i]\n", __FILE__, line);
		printf("[FAIL]\n");
		++NUM_TESTS_FAILED;
	} else if (!expectSuccess) {
		printf("Command not recognized; this is OK.\n");
		printf("[PASS]\n");
	}
	
	printQueue();
	printf("------------------------------------------\n");
}


///////////////////////////////////////////////////////////////////////////////
// sanity check on the queue
void checkQueue(int Expected_numElementsInQueue)
{
	if (wci_RXBUFFER.numElementsInQueue != Expected_numElementsInQueue) {
		printf("--->expected %i elements in queue, but got %i\n", Expected_numElementsInQueue, wci_RXBUFFER.numElementsInQueue);
		printQueue();
		printf("[FAIL]\n");
		++NUM_TESTS_FAILED;
		printf("------------------------------------------\n");
	}
}


///////////////////////////////////////////////////////////////////////////////
//
void checkSpeed(int expectedSpeed)
{
	if (cf.arg[2] != expectedSpeed) {
		printf("--->expected SPEED %i, but got %i\n", expectedSpeed, cf.arg[2]);
		printf("[FAIL]\n");
		++NUM_TESTS_FAILED;
		printf("------------------------------------------\n");
	}
}


///////////////////////////////////////////////////////////////////////////////
// main test driver
int main(int argc, char** argv)
{
	printf("TESTING wci module:\n");
	printf("-------------------\n");
	if (wci_RXBUFFER.size < 32) {
		printf("queue must be at least 32 bytes for these tests\n");
		printf("[FAIL]\n");		
		++NUM_TESTS_FAILED;
		goto end; // perfect use of 'goto'
	}

	runTest("$RST;", TRUE, CMD_OPCODE_RESET, __LINE__);
	runTest("$ReST;", FALSE, 0, __LINE__);
	runTest("$RST;", TRUE, CMD_OPCODE_RESET, __LINE__);	
	runTest("$ReST;$RST;", TRUE, CMD_OPCODE_RESET, __LINE__);
	
	// buffer should be empty at this point
	checkQueue(0);
	runTest("", FALSE, 0, __LINE__);	
	checkQueue(0);

	runTest("RST$RST$RST;", TRUE, CMD_OPCODE_RESET, __LINE__);
	checkQueue(0);
	runTest("", FALSE, 0, __LINE__);	
	checkQueue(0);

	runTest("$$$$$$$$$$$$$$$$RUN;", TRUE, CMD_OPCODE_RUN, __LINE__);
	runTest("$RUN$RUN;$$RU;$RUN;$", TRUE, CMD_OPCODE_RUN, __LINE__);
	runTest("", TRUE, CMD_OPCODE_RUN, __LINE__);
	checkQueue(1);  // $
	runTest("", FALSE, 0, __LINE__);
	checkQueue(1);  // $
	runTest("RUN", FALSE, 0, __LINE__);	
	checkQueue(4);  // $RUN
	runTest("", FALSE, 0, __LINE__);
	checkQueue(4);  // $RUN
	runTest(";", TRUE, CMD_OPCODE_RUN, __LINE__);
	checkQueue(0);  // $RUN

	runTest("$$$$$$RUNRST;RST$RST$RUN;;;;;$$", TRUE, CMD_OPCODE_RUN, __LINE__);
	runTest("", FALSE, 0, __LINE__);	
	checkQueue(1);  // $
	runTest(";", FALSE, 0, __LINE__);
	checkQueue(0);

	runTest("$STOP;", TRUE, CMD_OPCODE_STOP, __LINE__);
	runTest("$STP;$STOPP;S$TOP$$;;$STOP", FALSE, 0, __LINE__);
	runTest(";$RUN;$RST$ST;OP;;;$STOP;;", TRUE, CMD_OPCODE_STOP, __LINE__);
	runTest("", TRUE, CMD_OPCODE_RUN, __LINE__);
	runTest("", TRUE, CMD_OPCODE_STOP, __LINE__);
	checkQueue(1);  // ;

	runTest("$PING;", TRUE, CMD_OPCODE_PING, __LINE__);
	checkQueue(0);
	runTest("$PPING;PING;;;$PING$PING;$RST;", TRUE, CMD_OPCODE_PING, __LINE__);
	runTest("", TRUE, CMD_OPCODE_RESET, __LINE__);
	checkQueue(0);

	runTest("$MOT1>128;", TRUE, CMD_OPCODE_MOTOR, __LINE__);
	checkSpeed(128); 
	runTest("$MOT1<000;", TRUE, CMD_OPCODE_MOTOR, __LINE__);
	checkSpeed(0); 
	runTest("$MOT0>255;", TRUE, CMD_OPCODE_MOTOR, __LINE__);
	checkSpeed(255); 
	runTest("$MOT2^013;", TRUE, CMD_OPCODE_MOTOR, __LINE__);
	checkSpeed(13); 
	runTest("$MOT2v088;", TRUE, CMD_OPCODE_MOTOR, __LINE__);
	checkSpeed(88);  // 88 MILES AN HOUR!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

	runTest("$MOT0^123;", FALSE, 0, __LINE__); // motor 0 can't go up
	runTest("$MOT0v117;", FALSE, 0, __LINE__); // motor 0 can't go down
	runTest("$MOT1^123;", FALSE, 0, __LINE__); // motor 1 can't go up
	runTest("$MOT1v117;", FALSE, 0, __LINE__); // motor 1 can't go down

	runTest("$MOT2<083;", FALSE, 0, __LINE__); // up/down motor can't go left/right
	runTest("$MOT2>083;", FALSE, 0, __LINE__); // up/down motor can't go left/right

	runTest("$MOT0>355;", FALSE, 0, __LINE__); // speeding
	runTest("$MOT1<999;", FALSE, 0, __LINE__); // speeding
	runTest("$MOT2^256;", FALSE, 0, __LINE__); // speeding
	
	runTest("$MOT3>055;", FALSE, 0, __LINE__);
	runTest("$MOT3<055;", FALSE, 0, __LINE__);
	runTest("$MOT3^055;", FALSE, 0, __LINE__);
	runTest("$MOT3v055;", FALSE, 0, __LINE__);
	checkQueue(0);

	runTest(";$MOT2^257;;$$MOT0^000$MOT2v254;", TRUE, CMD_OPCODE_MOTOR, __LINE__);

	runTest("j28vhjGjgy984Y$PING;3jt#U$JGJdh3", TRUE, CMD_OPCODE_PING, __LINE__);
	
	runTest("j28vhjGjgy984YJ3#alEhaEhajsSJdh3", FALSE, 0, __LINE__);
	checkQueue(0);

	runTest(";;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;", FALSE, 0, __LINE__);
	checkQueue(0);

	runTest("$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$", FALSE, 0, __LINE__);
	checkQueue(1);

	runTest("$$$$s$$$$ping$$$$$$$$$$a$$$$$$$$", FALSE, 0, __LINE__);
	checkQueue(0);

	runTest(";;;;;RUN;A;;s;;RST;;;PING;;;;;;;", FALSE, 0, __LINE__);
	checkQueue(0);

	runTest("$$$$$;;;$$;$;$;$;$$$$$;;;;$$;;;;", FALSE, 0, __LINE__);
	checkQueue(0);

	runTest("$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$", FALSE, 0, __LINE__);
	checkQueue(1);

	runTest("12345678", FALSE, 0, __LINE__);
	checkQueue(9);

	runTest("0", FALSE, 0, __LINE__);
	checkQueue(10);

	runTest("1", FALSE, 0, __LINE__);
	checkQueue(0);

end:
	if (NUM_TESTS_FAILED > 0) {
		printf("%i tests FAILED\n", NUM_TESTS_FAILED);
	} else {
		printf("All tests PASSED\n");
	}

} // end main()