// Matthew Kraus
// wifi_ctrl_iface.c -- Wireless Control Interface
// module prefix: wci_
//
// Uses a state machine to parse incoming blimp commands from the network.
// Commands are coming from the PC and going to the blimp. This code
// will need to execute on the blimp and only understands commands targeted 
// for the blimp.

#include "wifi_ctrl_iface.h"

#include "../utils/queue.h"
#include "../utils/char.h"
#include "../utils/fsm.h"

// pc -> blimp commands
#define WCI_COMMAND_RESET "RST"
#define WCI_COMMAND_RUN   "RUN"
#define WCI_COMMAND_STOP  "STOP"
#define WCI_COMMAND_MOTOR "MOT"
#define WCI_COMMAND_QUERY "QRY"
#define WCI_COMMAND_PING  "PING"
#define WCI_COMMAND_PID   "PID"
#define WCI_COMMAND_CA    "CA"
#define WCI_COMMAND_TAR   "TAR"

#define WCI_BUFFER_SIZE 32
static QUEUE_DATATYPE WCI_BUFFER[WCI_BUFFER_SIZE];

// This is the maximum size that any command can be in bytes.
// This is needed in the event that the parser sees the start
// of a command '$', but doesn't see the end ';'. If it sees
// the start '$' but doesn't see the end, and there are more
// than WCI_MAX_COMMAND_LENGTH characters, it knows that it is
// an invalid command. WCI_BUFFER_SIZE must be larger than
// WCI_MAX_COMMAND_LENGTH
#define WCI_MAX_COMMAND_LENGTH 10

// TODO: ALL accesses to this must be protected w/ critical section!!!
QUEUE_DATA wci_RXBUFFER = {
	WCI_BUFFER,
	WCI_BUFFER_SIZE,
	0,
	0,
	0
}; 

// even if the network buffer fills up, we won't parse the next
// command until we've handed off the current command frame
static uint8_t isNextFrameReady = FALSE;

// this is built by each state's onEnter() method
static CMD_COMMAND_FRAME currentCommandFrame;


///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                 SECTION 1: STATE TRANSITION FUNCTIONS
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

// helper function
static void consumeBuffer(QUEUE_DATA* queue, uint8_t len)
{
	QUEUE_DATATYPE val;

	while (!queue_isEmpty(queue) && len > 0) {
		queue_pop(queue, &val);
		--len;
	}
}

///////////////////////////////////////////////////////////////////////////////
// RST -- parse reset command
///////////////////////////////////////////////////////////////////////////////
static BOOL wci_reset_acceptsInput(void* rawData)
{	
	QUEUE_DATATYPE val;
	QUEUE_DATA* pQueue;		
	
	pQueue = (QUEUE_DATA*)rawData;	
	val = 0;
		
	// check for end of frame marker
	queue_peekAt(pQueue, sizeof(WCI_COMMAND_RESET) / sizeof(char) - 1, &val);
	
	return queue_startsWith(pQueue, WCI_COMMAND_RESET) && val == ';';
}

static void wci_reset_onEnter(void* rawData)
{
	QUEUE_DATA* pQueue;
	QUEUE_DATATYPE val;	

	pQueue = (QUEUE_DATA*)rawData;	
	
	consumeBuffer(pQueue, sizeof(WCI_COMMAND_RESET) / sizeof(char) - 1);

	// check for end of frame marker & throw away (we should always have to do this)
	//if (queue_startsWith(pQueue, ";") ) queue_pop(pQueue, &val);
	queue_pop(pQueue, &val);

	// build RESET command frame
	currentCommandFrame.OPcode = CMD_OPCODE_RESET;  // no arguments	
	
	isNextFrameReady = TRUE; // frame is valid
}


///////////////////////////////////////////////////////////////////////////////
// RUN -- parse run command
///////////////////////////////////////////////////////////////////////////////
static BOOL wci_run_acceptsInput(void* rawData)
{
	QUEUE_DATATYPE val;
	QUEUE_DATA* pQueue;		
	
	pQueue = (QUEUE_DATA*)rawData;
	val = 0;
		
	// check for end of frame marker
	queue_peekAt(pQueue, sizeof(WCI_COMMAND_RUN) / sizeof(char) - 1, &val);
	
	return queue_startsWith(pQueue, WCI_COMMAND_RUN) && val == ';';
}

static void wci_run_onEnter(void* rawData)
{
	QUEUE_DATA* pQueue;
	QUEUE_DATATYPE val;	

	pQueue = (QUEUE_DATA*)rawData;	
	
	consumeBuffer(pQueue, sizeof(WCI_COMMAND_RUN) / sizeof(char) - 1);

	// check for end of frame marker & throw away (we should always have to do this)
	// if (queue_startsWith(pQueue, ";") ) queue_pop(pQueue, &val);
	queue_pop(pQueue, &val);

	// build RUN command frame
	currentCommandFrame.OPcode = CMD_OPCODE_RUN;  // no arguments	
	
	isNextFrameReady = TRUE; // frame is valid
}


///////////////////////////////////////////////////////////////////////////////
// STOP -- parse stop command
///////////////////////////////////////////////////////////////////////////////
static BOOL wci_stop_acceptsInput(void* rawData)
{
	QUEUE_DATATYPE val;
	QUEUE_DATA* pQueue;		
	
	pQueue = (QUEUE_DATA*)rawData;
	val = 0;

	// check for end of frame marker
	queue_peekAt(pQueue, sizeof(WCI_COMMAND_STOP) / sizeof(char) - 1, &val);
	
	return queue_startsWith(pQueue, WCI_COMMAND_STOP) && val == ';';
}

static void wci_stop_onEnter(void* rawData)
{
	QUEUE_DATA* pQueue;
	QUEUE_DATATYPE val;	

	pQueue = (QUEUE_DATA*)rawData;	
	
	consumeBuffer(pQueue, sizeof(WCI_COMMAND_STOP) / sizeof(char) - 1);

	// check for end of frame marker & throw away (we should always have to do this)
	//if (queue_startsWith(pQueue, ";") ) queue_pop(pQueue, &val);
	queue_pop(pQueue, &val);

	// build command frame
	currentCommandFrame.OPcode = CMD_OPCODE_STOP;  // no arguments	
	
	isNextFrameReady = TRUE; // frame is valid
}


///////////////////////////////////////////////////////////////////////////////
// PING -- keepalive, ping
///////////////////////////////////////////////////////////////////////////////
static BOOL wci_ping_acceptsInput(void* rawData)
{
	QUEUE_DATATYPE val;
	QUEUE_DATA* pQueue;		
	
	pQueue = (QUEUE_DATA*)rawData;
	val = 0;

	// check for end of frame marker
	queue_peekAt(pQueue, sizeof(WCI_COMMAND_PING) / sizeof(char) - 1, &val);
	
	return queue_startsWith(pQueue, WCI_COMMAND_PING) && val == ';';
}

static void wci_ping_onEnter(void* rawData)
{
	QUEUE_DATA* pQueue;
	QUEUE_DATATYPE val;	

	pQueue = (QUEUE_DATA*)rawData;	
	
	consumeBuffer(pQueue, sizeof(WCI_COMMAND_PING) / sizeof(char) - 1);

	// check for end of frame marker & throw away (we should always have to do this)
	//if (queue_startsWith(pQueue, ";") ) queue_pop(pQueue, &val);
	queue_pop(pQueue, &val);

	// build command frame
	currentCommandFrame.OPcode = CMD_OPCODE_PING;  // no arguments	
	
	isNextFrameReady = TRUE; // frame is valid
}


///////////////////////////////////////////////////////////////////////////////
// MOT -- motor command {"Motor ID[0,1,2]", "Direction [Fwd,Rev]", "Duty Cycle[0-255]"}
//
// Because this command takes arguments, STRICT validation is performed.
///////////////////////////////////////////////////////////////////////////////
static BOOL wci_motor_acceptsInput(void* rawData)
{
	QUEUE_DATATYPE val;
	QUEUE_DATATYPE d1;
	QUEUE_DATATYPE d2;
	QUEUE_DATATYPE d3;
	QUEUE_DATATYPE ID;
	QUEUE_DATATYPE DIR;
	QUEUE_DATA* pQueue;
	
	pQueue = (QUEUE_DATA*)rawData;
	val = 0;

	if (!queue_startsWith(pQueue, WCI_COMMAND_MOTOR) ) return FALSE;

	// check for end of frame marker
	queue_peekAt(pQueue, sizeof(WCI_COMMAND_MOTOR) / sizeof(char) - 1 + 5, &val);
	if (val != ';') return FALSE;

	// check ID
	queue_peekAt(pQueue, sizeof(WCI_COMMAND_MOTOR) / sizeof(char) - 1 + 0, &ID);
	if (ID != CMD_MOT_ID0 && ID != CMD_MOT_ID1 && ID != CMD_MOT_ID2) return FALSE;

	// check DIR
	queue_peekAt(pQueue, sizeof(WCI_COMMAND_MOTOR) / sizeof(char) - 1 + 1, &DIR);
	if (ID == CMD_MOT_ID0 || ID == CMD_MOT_ID1) {
		// left/right fans can only go fwd/rev
		if (DIR != CMD_MOT_DIR_FWD && DIR != CMD_MOT_DIR_REV) return FALSE;
	}
	else if (ID == CMD_MOT_ID2) {
		// only up/down fan can go up/down
		if (DIR != CMD_MOT_DIR_UP && DIR != CMD_MOT_DIR_DOWN) return FALSE;
	}

	// check SPEED
	queue_peekAt(pQueue, sizeof(WCI_COMMAND_MOTOR) / sizeof(char) - 1 + 2, &d1);
	if (d1 < 48 || d1 > 50) return FALSE;  // 0,1,2
	queue_peekAt(pQueue, sizeof(WCI_COMMAND_MOTOR) / sizeof(char) - 1 + 3, &d2);
	if (d2 < 48 || d2 > 57) return FALSE;  // 0-9
	if (d1 == 50 && d2 > 53) return FALSE; // > 260
	queue_peekAt(pQueue, sizeof(WCI_COMMAND_MOTOR) / sizeof(char) - 1 + 4, &d3);
	if (d3 < 48 || d3 > 57) return FALSE;  // 0-9
	if (d1 == 50 && d2 == 53 && d3 > 53) return FALSE;

	return TRUE;
}

static void wci_motor_onEnter(void* rawData)
{
	QUEUE_DATA* pQueue;
	QUEUE_DATATYPE val;	
	uint8_t speed;

	pQueue = (QUEUE_DATA*)rawData;	
	
	consumeBuffer(pQueue, sizeof(WCI_COMMAND_MOTOR) / sizeof(char) - 1);

	// build command frame
	currentCommandFrame.OPcode = CMD_OPCODE_MOTOR;  // no arguments	

	// get ID
	queue_pop(pQueue, &val);
	currentCommandFrame.arg[0] = val;

	// DIR
	queue_pop(pQueue, &val);
	currentCommandFrame.arg[1] = val;

	// SPEED
	speed = 0;
	queue_pop(pQueue, &val);
	val = val - 48; // convert ASCII to int
	speed += (val << 6) + (val << 5) + (val << 2); // val * 100
	
	queue_pop(pQueue, &val);
	val = val - 48; // convert ASCII to int
	speed += (val << 3) + (val << 1); // val * 10
	
	queue_pop(pQueue, &val);
	val = val - 48; // convert ASCII to int
	speed += val;

	currentCommandFrame.arg[2] = speed;

	// check for end of frame marker & throw away (we should always have to do this)
	//if (queue_startsWith(pQueue, ";") ) queue_pop(pQueue, &val);
	queue_pop(pQueue, &val);

	isNextFrameReady = TRUE; // frame is valid
}


///////////////////////////////////////////////////////////////////////////////
// PID -- turn on/off PID
///////////////////////////////////////////////////////////////////////////////
static BOOL wci_pid_acceptsInput(void* rawData)
{
	QUEUE_DATATYPE val;
	QUEUE_DATATYPE d1;
	QUEUE_DATA* pQueue;

	pQueue = (QUEUE_DATA*)rawData;
	val = 0;

	if (!queue_startsWith(pQueue, WCI_COMMAND_PID) ) return FALSE;

	// check for end of frame marker
	queue_peekAt(pQueue, sizeof(WCI_COMMAND_PID) / sizeof(char) - 1 + 1, &val);
	if (val != ';') return FALSE;

	// check on/off
	queue_peekAt(pQueue, sizeof(WCI_COMMAND_PID) / sizeof(char) - 1 + 0, &d1);
	if (d1 != '0' && d1 != '1') {
		return FALSE;  // 0,1
	}

	return TRUE;
}

static void wci_pid_onEnter(void* rawData)
{
	QUEUE_DATA* pQueue;
	QUEUE_DATATYPE val;

	pQueue = (QUEUE_DATA*)rawData;

	consumeBuffer(pQueue, sizeof(WCI_COMMAND_PID) / sizeof(char) - 1);

	// build command frame
	currentCommandFrame.OPcode = CMD_OPCODE_PID;  // no arguments

	// get on/off
	queue_pop(pQueue, &val);
	currentCommandFrame.arg[0] = val - 48; // convert ASCII to int;

	// throw away end of frame marker
	queue_pop(pQueue, &val);

	isNextFrameReady = TRUE; // frame is valid
}


///////////////////////////////////////////////////////////////////////////////
// CA -- turn on/off collison avoidance
///////////////////////////////////////////////////////////////////////////////
static BOOL wci_ca_acceptsInput(void* rawData)
{
	QUEUE_DATATYPE val;
	QUEUE_DATATYPE d1;
	QUEUE_DATA* pQueue;

	pQueue = (QUEUE_DATA*)rawData;
	val = 0;

	if (!queue_startsWith(pQueue, WCI_COMMAND_CA) ) return FALSE;

	// check for end of frame marker
	queue_peekAt(pQueue, sizeof(WCI_COMMAND_CA) / sizeof(char) - 1 + 1, &val);
	if (val != ';') return FALSE;

	// check on/off
	queue_peekAt(pQueue, sizeof(WCI_COMMAND_CA) / sizeof(char) - 1 + 0, &d1);
	if (d1 != '0' && d1 != '1') {
		return FALSE;  // 0,1
	}

	return TRUE;
}

static void wci_ca_onEnter(void* rawData)
{
	QUEUE_DATA* pQueue;
	QUEUE_DATATYPE val;

	pQueue = (QUEUE_DATA*)rawData;

	consumeBuffer(pQueue, sizeof(WCI_COMMAND_CA) / sizeof(char) - 1);

	// build command frame
	currentCommandFrame.OPcode = CMD_OPCODE_CA;  // no arguments

	// get on/off
	queue_pop(pQueue, &val);
	currentCommandFrame.arg[0] = val - 48; // convert ASCII to int;

	// throw away end of frame marker
	queue_pop(pQueue, &val);

	isNextFrameReady = TRUE; // frame is valid
}


///////////////////////////////////////////////////////////////////////////////
// TAR -- TARget altitude
///////////////////////////////////////////////////////////////////////////////
static BOOL wci_tar_acceptsInput(void* rawData)
{
	QUEUE_DATATYPE val;
	QUEUE_DATATYPE d1;
	QUEUE_DATATYPE d2;
	QUEUE_DATA* pQueue;

	pQueue = (QUEUE_DATA*)rawData;
	val = 0;

	if (!queue_startsWith(pQueue, WCI_COMMAND_TAR) ) return FALSE;

	// check for end of frame marker
	queue_peekAt(pQueue, sizeof(WCI_COMMAND_TAR) / sizeof(char) - 1 + 2, &val);
	if (val != ';') return FALSE;

	// check altitude value
	queue_peekAt(pQueue, sizeof(WCI_COMMAND_TAR) / sizeof(char) - 1 + 0, &d1);
	if (d1 < 48 || d1 > 57) return FALSE;  // 0-9
	queue_peekAt(pQueue, sizeof(WCI_COMMAND_TAR) / sizeof(char) - 1 + 1, &d2);
	if (d2 < 48 || d2 > 57) return FALSE;  // 0-9

	return TRUE;
}

static void wci_tar_onEnter(void* rawData)
{
	QUEUE_DATA* pQueue;
	QUEUE_DATATYPE val;
	uint8_t alt;

	pQueue = (QUEUE_DATA*)rawData;

	consumeBuffer(pQueue, sizeof(WCI_COMMAND_TAR) / sizeof(char) - 1);

	// build command frame
	currentCommandFrame.OPcode = CMD_OPCODE_TAR;  // no arguments

	// altitude
	alt = 0;
	queue_pop(pQueue, &val);
	val = val - 48; // convert ASCII to int
	alt += (val << 3) + (val << 1); // val * 10

	queue_pop(pQueue, &val);
	val = val - 48; // convert ASCII to int
	alt += val;

	currentCommandFrame.arg[0] = alt;

	// check for end of frame marker & throw away (we should always have to do this)
	queue_pop(pQueue, &val);

	isNextFrameReady = TRUE; // frame is valid
}

///////////////////////////////////////////////////////////////////////////////
// PARSE state--used for finding $ frame markers and getting back in sync
// if we get corrupted data over the network or if commands are invalid.
// The parse state should be the default state in most cases.
///////////////////////////////////////////////////////////////////////////////
static BOOL wci_parseState_acceptsInput(void* rawData)
{
	// we accept any input
	return TRUE;
}

static void wci_parseState_onEnter(void* rawData)
{
	PQUEUE_DATA queue = (PQUEUE_DATA)rawData;
	QUEUE_DATATYPE val;

	
	/*		
	while (queue_contains(queue, ';') ) {
		if (queue_startsWith(queue, "$")) {			
			while(queue_peekAt(queue, 0, &val) && val == '$') {
				// get in sync with start of frame marker (discard all of them)
				queue_pop(queue, &val);
			}
			return; // we are in sync now

		} else {
			// try to get back in sync
			queue_pop(queue, &val);
		}
	} // end while
	*/
}


///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                 SECTION 2: STATE MACHINE DEFINITION
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

//forwards
static FSM_STATE wci_parseState;  

//
// RUN command ////////////////
//
FSM_STATE wci_state_run = {
	wci_run_acceptsInput,
	NULL,
	0,
	&wci_parseState,
	wci_run_onEnter,
	NULL
};


//
// RESET command //////////////
//
FSM_STATE wci_state_reset = {	
	wci_reset_acceptsInput,	
	NULL,	
	0,
	&wci_parseState,
	wci_reset_onEnter,
	NULL
};

//
// STOP command
//
FSM_STATE wci_state_stop = {
	wci_stop_acceptsInput,
	NULL,
	0,
	&wci_parseState,
	wci_stop_onEnter,
	NULL
};


//
// PING command
//
FSM_STATE wci_state_ping = {
	wci_ping_acceptsInput,
	NULL,
	0,
	&wci_parseState,
	wci_ping_onEnter,
	NULL
};


//
// MOT command
//
FSM_STATE wci_state_motor = {
	wci_motor_acceptsInput,
	NULL,
	0,
	&wci_parseState,
	wci_motor_onEnter,
	NULL
};

//
// PID command
//
FSM_STATE wci_state_pid = {
	wci_pid_acceptsInput,
	NULL,
	0,
	&wci_parseState,
	wci_pid_onEnter,
	NULL
};


//
// CA command
//
FSM_STATE wci_state_ca = {
	wci_ca_acceptsInput,
	NULL,
	0,
	&wci_parseState,
	wci_ca_onEnter,
	NULL
};


//
// TAR command
//
FSM_STATE wci_state_tar = {
	wci_tar_acceptsInput,
	NULL,
	0,
	&wci_parseState,
	wci_tar_onEnter,
	NULL
};

//
// PARSE (default) state //////
//
static FSM_STATE* wci_parseState_nextStateList[] = {
	// when adding a state to this list, update numStates in wci_parseState also!!!
	&wci_state_motor, &wci_state_ping, &wci_state_reset, &wci_state_run, &wci_state_stop, &wci_state_pid, &wci_state_ca, &wci_state_tar
};
FSM_STATE wci_parseState = {
	wci_parseState_acceptsInput,
	wci_parseState_nextStateList,
	8, // update me if you add a state to the state list!!!
	&wci_parseState,
	wci_parseState_onEnter,
	NULL
};

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                 SECTION 3: DRIVER METHODS -- run the state machine
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

// Starting state
static FSM_STATE* currentState = &wci_parseState;

//
// This section constitutes the public interface for the wci module.
// Methods outside of this module should only call these functions
// and should not directly call functions in section 1.
//

///////////////////////////////////////////////////////////////////////////////
// Returns true if a command frame is ready. If a command frame is not already
// ready, then it tries to process new data to prepare a full command frame.
// If a full command frame can be formed, the command frame is prepared and
// true is returned. The new command frame can be retrieved by calling
// wci_nextCommand().
BOOL wci_hasNext()
{
	QUEUE_DATATYPE val;

	// don't process if we're already ready
	if (isNextFrameReady == TRUE) return TRUE;

	// TODO: BEGIN CRITICAL SECTION
	{
		// TODO: call uart_to_wci_copy()
				
		// consume buffer until we find a frame or exhaust buffer		
		while(!queue_isEmpty(&wci_RXBUFFER) ) {
			while(queue_peekAt(&wci_RXBUFFER, 0, &val) && val != '$') {
				queue_pop(&wci_RXBUFFER, &val); // sync to start of frame marker
			}
			while(queue_peekAt(&wci_RXBUFFER, 1, &val) && val == '$') {
				queue_pop(&wci_RXBUFFER, &val); // remove repeated $
			}
			
			// at this point, we are at $, or the queue is empty
			if (queue_peekAt(&wci_RXBUFFER, 0, &val) && val != '$') return FALSE;

			if (!queue_contains(&wci_RXBUFFER, ';') && wci_RXBUFFER.numElementsInQueue > WCI_MAX_COMMAND_LENGTH) {
				// no command is this long, so clear the buffer
				while(!queue_isEmpty(&wci_RXBUFFER) ) queue_pop(&wci_RXBUFFER, &val);  // TODO: write queue_clear()
				return FALSE;
			}
						
			if (queue_contains(&wci_RXBUFFER, ';') ) {
				// we saw start and end of frame markers, so we should have a full frame to process
				queue_pop(&wci_RXBUFFER, &val);            // prepare for parsing by removing $				
				fsm_onInput(currentState, &wci_RXBUFFER);  // TODO: synchronize access to buffer

				if (isNextFrameReady == TRUE) return TRUE;
			}
			else return FALSE;
		} // end while
	}
	// TODO: END CRITICAL SECTION

	return FALSE;
}


///////////////////////////////////////////////////////////////////////////////
// You *must* call wci_hasNext() and make sure TRUE is returned before calling
// this method. Failure to check will result in an invalid command frame.
void wci_nextCommand(CMD_COMMAND_FRAME* cf)
{
	// prepare to build the next command frame (after we return the current one)
	isNextFrameReady = FALSE;

	// return a copy of the current command frame	
	*cf = currentCommandFrame;
}

